from unittest.mock import patch

from django.test import TestCase, override_settings

from blueprints import tasks

from . import create_owner
from .testdata.load_entities import load_entities
from .testdata.load_eveuniverse import load_eveuniverse
from .testdata.load_locations import load_locations

TASKS_PATH = "blueprints.tasks"


@override_settings(CELERY_ALWAYS_EAGER=True)
class TestUpdateBlueprints(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_entities()
        load_eveuniverse()
        load_locations()
        cls.owner = create_owner(character_id=1101, corporation_id=2101)

    @patch(TASKS_PATH + ".Owner.update_blueprints_esi")
    def test_update_blueprints_for_owner(self, mock_update_blueprints_esi):
        tasks.update_blueprints_for_owner(self.owner.pk)
        self.assertTrue(mock_update_blueprints_esi.called)

    @patch(TASKS_PATH + ".Owner.update_blueprints_esi")
    def test_update_all_blueprints(self, mock_update_blueprints_esi):
        tasks.update_all_blueprints()
        self.assertTrue(mock_update_blueprints_esi.called)



from django.test import TestCase
from esi.errors import TokenError
from esi.models import Token

from blueprints.decorators import fetch_token_for_owner

from . import create_owner, scope_names_set
from .testdata.load_entities import load_entities

DUMMY_URL = "http://www.example.com"


class TestFetchToken(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_entities()

    def setUp(self) -> None:
        self.owner = create_owner(character_id=1101, corporation_id=2101)

    def test_specified_scope(self):
        @fetch_token_for_owner("esi-corporations.read_blueprints.v1")
        def dummy(owner, token):
            self.assertIsInstance(token, Token)
            self.assertIn("esi-corporations.read_blueprints.v1", scope_names_set(token))

        dummy(self.owner)

    def test_exceptions_if_not_found(self):
        @fetch_token_for_owner("invalid_scope")
        def dummy(owner, token):
            pass

        with self.assertRaises(TokenError):
            dummy(self.owner)



# Generated by Django 3.1.2 on 2020-12-24 15:08

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import blueprints.validators


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("eveuniverse", "0004_effect_longer_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("eveonline", "0012_index_additions"),
        ("authentication", "0017_remove_fleetup_permission"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app"),
                    ("request_blueprints", "Can request blueprints"),
                    ("manage_requests", "Can review and accept blueprint requests"),
                    (
                        "add_personal_blueprint_owner",
                        "Can add personal blueprint owners",
                    ),
                    (
                        "add_corporate_blueprint_owner",
                        "Can add corporate blueprint owners",
                    ),
                    ("view_alliance_blueprints", "Can view alliance's blueprints"),
                    (
                        "view_industry_jobs",
                        "Can view details about running industry jobs",
                    ),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Blueprint",
            fields=[
                (
                    "item_id",
                    models.PositiveBigIntegerField(
                        help_text="The EVE Item ID of the blueprint",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "location_flag",
                    models.CharField(
                        choices=[
                            ("AssetSafety", "AssetSafety"),
                            ("AutoFit", "AutoFit"),
                            ("Bonus", "Bonus"),
                            ("Booster", "Booster"),
                            ("BoosterBay", "BoosterBay"),
                            ("Capsule", "Capsule"),
                            ("Cargo", "Cargo"),
                            ("CorpDeliveries", "CorpDeliveries"),
                            ("CorpSAG1", "CorpSAG1"),
                            ("CorpSAG2", "CorpSAG2"),
                            ("CorpSAG3", "CorpSAG3"),
                            ("CorpSAG4", "CorpSAG4"),
                            ("CorpSAG5", "CorpSAG5"),
                            ("CorpSAG6", "CorpSAG6"),
                            ("CorpSAG7", "CorpSAG7"),
                            ("CrateLoot", "CrateLoot"),
                            ("Deliveries", "Deliveries"),
                            ("DroneBay", "DroneBay"),
                            ("DustBattle", "DustBattle"),
                            ("DustDatabank", "DustDatabank"),
                            ("FighterBay", "FighterBay"),
                            ("FighterTube0", "FighterTube0"),
                            ("FighterTube1", "FighterTube1"),
                            ("FighterTube2", "FighterTube2"),
                            ("FighterTube3", "FighterTube3"),
                            ("FighterTube4", "FighterTube4"),
                            ("FleetHangar", "FleetHangar"),
                            ("FrigateEscapeBay", "FrigateEscapeBay"),
                            ("Hangar", "Hangar"),
                            ("HangarAll", "HangarAll"),
                            ("HiSlot0", "HiSlot0"),
                            ("HiSlot1", "HiSlot1"),
                            ("HiSlot2", "HiSlot2"),
                            ("HiSlot3", "HiSlot3"),
                            ("HiSlot4", "HiSlot4"),
                            ("HiSlot5", "HiSlot5"),
                            ("HiSlot6", "HiSlot6"),
                            ("HiSlot7", "HiSlot7"),
                            ("HiddenModifiers", "HiddenModifiers"),
                            ("Implant", "Implant"),
                            ("Impounded", "Impounded"),
                            ("JunkyardReprocessed", "JunkyardReprocessed"),
                            ("JunkyardTrashed", "JunkyardTrashed"),
                            ("LoSlot0", "LoSlot0"),
                            ("LoSlot1", "LoSlot1"),
                            ("LoSlot2", "LoSlot2"),
                            ("LoSlot3", "LoSlot3"),
                            ("LoSlot4", "LoSlot4"),
                            ("LoSlot5", "LoSlot5"),
                            ("LoSlot6", "LoSlot6"),
                            ("LoSlot7", "LoSlot7"),
                            ("Locked", "Locked"),
                            ("MedSlot0", "MedSlot0"),
                            ("MedSlot1", "MedSlot1"),
                            ("MedSlot2", "MedSlot2"),
                            ("MedSlot3", "MedSlot3"),
                            ("MedSlot4", "MedSlot4"),
                            ("MedSlot5", "MedSlot5"),
                            ("MedSlot6", "MedSlot6"),
                            ("MedSlot7", "MedSlot7"),
                            ("OfficeFolder", "OfficeFolder"),
                            ("Pilot", "Pilot"),
                            ("PlanetSurface", "PlanetSurface"),
                            ("QuafeBay", "QuafeBay"),
                            ("QuantumCoreRoom", "QuantumCoreRoom"),
                            ("Reward", "Reward"),
                            ("RigSlot0", "RigSlot0"),
                            ("RigSlot1", "RigSlot1"),
                            ("RigSlot2", "RigSlot2"),
                            ("RigSlot3", "RigSlot3"),
                            ("RigSlot4", "RigSlot4"),
                            ("RigSlot5", "RigSlot5"),
                            ("RigSlot6", "RigSlot6"),
                            ("RigSlot7", "RigSlot7"),
                            ("SecondaryStorage", "SecondaryStorage"),
                            ("ServiceSlot0", "ServiceSlot0"),
                            ("ServiceSlot1", "ServiceSlot1"),
                            ("ServiceSlot2", "ServiceSlot2"),
                            ("ServiceSlot3", "ServiceSlot3"),
                            ("ServiceSlot4", "ServiceSlot4"),
                            ("ServiceSlot5", "ServiceSlot5"),
                            ("ServiceSlot6", "ServiceSlot6"),
                            ("ServiceSlot7", "ServiceSlot7"),
                            ("ShipHangar", "ShipHangar"),
                            ("ShipOffline", "ShipOffline"),
                            ("Skill", "Skill"),
                            ("SkillInTraining", "SkillInTraining"),
                            ("SpecializedAmmoHold", "SpecializedAmmoHold"),
                            (
                                "SpecializedCommandCenterHold",
                                "SpecializedCommandCenterHold",
                            ),
                            ("SpecializedFuelBay", "SpecializedFuelBay"),
                            ("SpecializedGasHold", "SpecializedGasHold"),
                            (
                                "SpecializedIndustrialShipHold",
                                "SpecializedIndustrialShipHold",
                            ),
                            ("SpecializedLargeShipHold", "SpecializedLargeShipHold"),
                            ("SpecializedMaterialBay", "SpecializedMaterialBay"),
                            ("SpecializedMediumShipHold", "SpecializedMediumShipHold"),
                            ("SpecializedMineralHold", "SpecializedMineralHold"),
                            ("SpecializedOreHold", "SpecializedOreHold"),
                            (
                                "SpecializedPlanetaryCommoditiesHold",
                                "SpecializedPlanetaryCommoditiesHold",
                            ),
                            ("SpecializedSalvageHold", "SpecializedSalvageHold"),
                            ("SpecializedShipHold", "SpecializedShipHold"),
                            ("SpecializedSmallShipHold", "SpecializedSmallShipHold"),
                            ("StructureActive", "StructureActive"),
                            ("StructureFuel", "StructureFuel"),
                            ("StructureInactive", "StructureInactive"),
                            ("StructureOffline", "StructureOffline"),
                            ("SubSystemBay", "SubSystemBay"),
                            ("SubSystemSlot0", "SubSystemSlot0"),
                            ("SubSystemSlot1", "SubSystemSlot1"),
                            ("SubSystemSlot2", "SubSystemSlot2"),
                            ("SubSystemSlot3", "SubSystemSlot3"),
                            ("SubSystemSlot4", "SubSystemSlot4"),
                            ("SubSystemSlot5", "SubSystemSlot5"),
                            ("SubSystemSlot6", "SubSystemSlot6"),
                            ("SubSystemSlot7", "SubSystemSlot7"),
                            ("Unlocked", "Unlocked"),
                            ("Wallet", "Wallet"),
                            ("Wardrobe", "Wardrobe"),
                        ],
                        help_text="Additional location information",
                        max_length=36,
                    ),
                ),
                (
                    "quantity",
                    models.PositiveIntegerField(
                        default=1, help_text="Number of blueprints"
                    ),
                ),
                (
                    "runs",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text="Runs remaining or null if the blueprint is an original",
                        null=True,
                    ),
                ),
                (
                    "material_efficiency",
                    models.PositiveIntegerField(
                        help_text="Material efficiency of the blueprint",
                        validators=[blueprints.validators.validate_material_efficiency],
                    ),
                ),
                (
                    "time_efficiency",
                    models.PositiveIntegerField(
                        help_text="Time efficiency of the blueprint",
                        validators=[blueprints.validators.validate_time_efficiency],
                    ),
                ),
                (
                    "eve_type",
                    models.ForeignKey(
                        help_text="Blueprint type",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Request",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "runs",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text="Runs requested or blank for maximum allowed",
                        null=True,
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("OP", "Open"),
                            ("IP", "In Progress"),
                            ("FL", "Fulfilled"),
                            ("CL", "Cancelled"),
                        ],
                        help_text="Status of the blueprint request",
                        max_length=2,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                (
                    "closed_at",
                    models.DateTimeField(blank=True, db_index=True, null=True),
                ),
                (
                    "blueprint",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="blueprints.blueprint",
                    ),
                ),
                (
                    "fulfulling_user",
                    models.ForeignKey(
                        blank=True,
                        help_text="The user that fulfilled the request, if it has been fulfilled",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "requesting_user",
                    models.ForeignKey(
                        help_text="The requesting user",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Owner",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="whether this owner is currently included in the sync process",
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="character used for syncing blueprints",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="authentication.characterownership",
                    ),
                ),
                (
                    "corporation",
                    models.OneToOneField(
                        blank=True,
                        default=None,
                        help_text="Corporation owning blueprints, if this is a 'corporate' owner",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveonline.evecorporationinfo",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Location",
            fields=[
                (
                    "id",
                    models.PositiveBigIntegerField(
                        help_text="Eve Online location ID, either item ID for stations or structure ID for structures",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="In-game name of this station or structure",
                        max_length=100,
                    ),
                ),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "eve_solar_system",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.evesolarsystem",
                    ),
                ),
                (
                    "eve_type",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.evetype",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="corporation this station or structure belongs to",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="Eve Online location ID of the parent object",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="blueprints.location",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="IndustryJob",
            fields=[
                (
                    "id",
                    models.PositiveBigIntegerField(
                        help_text="Eve Online job ID", primary_key=True, serialize=False
                    ),
                ),
                (
                    "activity",
                    models.PositiveIntegerField(
                        choices=[
                            (1, "Manufacturing"),
                            (2, "Researching Technology"),
                            (3, "Researching Time Efficiency"),
                            (4, "Researching Material Efficiency"),
                            (5, "Copying"),
                            (6, "Duplicating"),
                            (7, "Reverse Engineering"),
                            (8, "Inventing"),
                            (9, "Reacting"),
                        ]
                    ),
                ),
                ("start_date", models.DateTimeField()),
                ("end_date", models.DateTimeField()),
                ("runs", models.PositiveIntegerField()),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("active", "Active"),
                            ("cancelled", "Cancelled"),
                            ("delivered", "Delivered"),
                            ("paused", "Paused"),
                            ("ready", "Ready"),
                            ("reverted", "Reverted"),
                        ],
                        max_length=10,
                    ),
                ),
                (
                    "blueprint",
                    models.OneToOneField(
                        help_text="Blueprint the job is running",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="blueprints.blueprint",
                    ),
                ),
                (
                    "installer",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="jobs",
                        to="eveonline.evecharacter",
                    ),
                ),
                (
                    "location",
                    models.ForeignKey(
                        help_text="Eve Online location ID of the facility in which the job is running",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="blueprints.location",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="jobs",
                        to="blueprints.owner",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="blueprint",
            name="location",
            field=models.ForeignKey(
                help_text="Blueprint location",
                on_delete=django.db.models.deletion.CASCADE,
                to="blueprints.location",
            ),
        ),
        migrations.AddField(
            model_name="blueprint",
            name="owner",
            field=models.ForeignKey(
                help_text="Corporation that owns the blueprint",
                on_delete=django.db.models.deletion.CASCADE,
                to="blueprints.owner",
            ),
        ),
    ]



"""Models for Blueprints."""

from typing import List, Tuple

from django.contrib.auth.models import Permission, User
from django.db import models
from django.utils.translation import gettext_lazy as _
from esi.errors import TokenExpiredError, TokenInvalidError
from esi.models import Token
from eveuniverse.models import EveEntity, EveSolarSystem, EveType

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.evelinks import dotlan
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger
from app_utils.django import users_with_permission
from app_utils.logging import LoggerAddTag, make_logger_prefix

from . import __title__
from .constants import EVE_LOCATION_FLAGS
from .decorators import fetch_token_for_owner
from .managers import BlueprintManager, LocationManager, RequestManager
from .providers import esi
from .validators import validate_material_efficiency, validate_time_efficiency

NAMES_MAX_LENGTH = 100

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class General(models.Model):
    """Meta model for app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app"),
            ("request_blueprints", "Can request blueprints"),
            ("manage_requests", "Can review and accept blueprint requests"),
            ("add_personal_blueprint_owner", "Can add personal blueprint owners"),
            ("add_corporate_blueprint_owner", "Can add corporate blueprint owners"),
            ("view_alliance_blueprints", "Can view alliance's blueprints"),
            ("view_blueprint_locations", "Can view the location of all blueprints"),
            ("view_industry_jobs", "Can view details about running industry jobs"),
        )


class Owner(models.Model):
    """A corporation that owns blueprints"""

    ERROR_NONE = 0
    ERROR_TOKEN_INVALID = 1
    ERROR_TOKEN_EXPIRED = 2
    ERROR_INSUFFICIENT_PERMISSIONS = 3
    ERROR_NO_CHARACTER = 4
    ERROR_ESI_UNAVAILABLE = 5
    ERROR_OPERATION_MODE_MISMATCH = 6
    ERROR_UNKNOWN = 99

    ERRORS_LIST = [
        (ERROR_NONE, "No error"),
        (ERROR_TOKEN_INVALID, "Invalid token"),
        (ERROR_TOKEN_EXPIRED, "Expired token"),
        (ERROR_INSUFFICIENT_PERMISSIONS, "Insufficient permissions"),
        (ERROR_NO_CHARACTER, "No character set for fetching data from ESI"),
        (ERROR_ESI_UNAVAILABLE, "ESI API is currently unavailable"),
        (
            ERROR_OPERATION_MODE_MISMATCH,
            "Operaton mode does not match with current setting",
        ),
        (ERROR_UNKNOWN, "Unknown error"),
    ]

    corporation = models.OneToOneField(
        EveCorporationInfo,
        default=None,
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        help_text="Corporation owning blueprints, if this is a 'corporate' owner",
        related_name="+",
    )
    character = models.ForeignKey(
        CharacterOwnership,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        help_text="character used for syncing blueprints",
        related_name="+",
    )
    is_active = models.BooleanField(
        default=True,
        help_text=("whether this owner is currently included in the sync process"),
    )

    class Meta:
        default_permissions = ()

    def __str__(self):
        return self.name

    @property
    def name(self) -> str:
        try:
            if self.corporation:
                return self.corporation.corporation_name
            return self.character.character.character_name
        except AttributeError:
            return ""

    def update_locations_esi(self):
        if self.corporation:
            assets = self._fetch_corporate_assets()
            token = self.token(
                [
                    "esi-universe.read_structures.v1",
                    "esi-assets.read_corporation_assets.v1",
                ]
            )[0]
        else:
            assets = self._fetch_personal_assets()
            token = self.token(
                ["esi-universe.read_structures.v1", "esi-assets.read_assets.v1"]
            )[0]

        asset_ids = []
        asset_locations = {}
        assets_by_id = {}
        for asset in assets:
            asset_ids.append(asset["item_id"])
            assets_by_id[asset["item_id"]] = asset

        for asset in assets:
            if asset["location_id"] in asset_ids:
                location_id = asset["location_id"]
                if asset_locations.get(location_id):
                    asset_locations[location_id].append(asset["item_id"])
                else:
                    asset_locations[location_id] = [asset["item_id"]]

        for location in list(asset_locations.keys()):
            if not Location.objects.filter(id=location).count() > 0:
                parent_location = assets_by_id[location]["location_id"]
                Location.objects.create(
                    id=location,
                    parent=self._fetch_location(parent_location, token=token),
                    eve_type=EveType.objects.filter(
                        id=assets_by_id[location]["type_id"]
                    ).first(),
                )
            else:
                parent_location = assets_by_id[location]["location_id"]
                location_obj = Location.objects.filter(id=location).first()
                if location_obj:
                    location_obj.parent = self._fetch_location(
                        parent_location, token=token
                    )
                    location_obj.save()

    def update_blueprints_esi(self):
        """updates all blueprints from ESI"""

        if self.is_active:
            blueprint_ids_to_remove = list(
                Blueprint.objects.filter(owner=self).values_list("item_id", flat=True)
            )
            if self.corporation:
                blueprints = self._fetch_corporate_blueprints()
                token = self.token(
                    [
                        "esi-universe.read_structures.v1",
                        "esi-corporations.read_blueprints.v1",
                    ]
                )[0]
            else:
                blueprints = self._fetch_personal_blueprints()
                token = self.token(
                    [
                        "esi-universe.read_structures.v1",
                        "esi-characters.read_blueprints.v1",
                    ]
                )[0]

            for blueprint in blueprints:
                runs = blueprint["runs"]
                if runs < 1:
                    runs = None
                quantity = blueprint["quantity"]
                if quantity < 0:
                    quantity = 1
                original = Blueprint.objects.filter(
                    owner=self, item_id=blueprint["item_id"]
                ).first()
                if original is not None:
                    # We've seen this blueprint coming from ESI, so we know it shouldn't be deleted
                    blueprint_ids_to_remove.remove(original.item_id)
                    original.location = self._fetch_location(
                        blueprint["location_id"],
                        token=token,
                    )
                    original.location_flag = blueprint["location_flag"]
                    original.eve_type = EveType.objects.get_or_create_esi(
                        id=blueprint["type_id"]
                    )[0]
                    original.runs = runs
                    original.material_efficiency = blueprint["material_efficiency"]
                    original.time_efficiency = blueprint["time_efficiency"]
                    original.quantity = quantity
                    original.save()
                else:
                    Blueprint.objects.create(
                        owner=self,
                        location=self._fetch_location(
                            blueprint["location_id"],
                            token=token,
                        ),
                        location_flag=blueprint["location_flag"],
                        eve_type=EveType.objects.get_or_create_esi(
                            id=blueprint["type_id"]
                        )[0],
                        item_id=blueprint["item_id"],
                        runs=runs,
                        material_efficiency=blueprint["material_efficiency"],
                        time_efficiency=blueprint["time_efficiency"],
                        quantity=quantity,
                    )
            Blueprint.objects.filter(pk__in=blueprint_ids_to_remove).delete()

    def update_industry_jobs_esi(self):
        """updates all blueprints from ESI"""

        if self.is_active:
            job_ids_to_remove = list(
                IndustryJob.objects.filter(owner=self).values_list("id", flat=True)
            )
            if self.corporation:
                jobs = self._fetch_corporate_industry_jobs()
                token = self.token(
                    [
                        "esi-universe.read_structures.v1",
                        "esi-industry.read_corporation_jobs.v1",
                    ]
                )[0]
            else:
                jobs = self._fetch_personal_industry_jobs()
                token = self.token(
                    [
                        "esi-universe.read_structures.v1",
                        "esi-industry.read_character_jobs.v1",
                    ]
                )[0]

            for job in jobs:
                original = IndustryJob.objects.filter(
                    owner=self, id=job["job_id"]
                ).first()
                blueprint = Blueprint.objects.filter(pk=job["blueprint_id"]).first()
                if blueprint is not None:
                    if original is not None:
                        # We've seen this job coming from ESI, so we know it shouldn't be deleted
                        job_ids_to_remove.remove(original.id)
                        original.status = job["status"]
                        original.save()
                    else:
                        # Reject personal listings of corporate jobs and visa-versa
                        if blueprint.owner == self:
                            installer = EveCharacter.objects.get_character_by_id(
                                job["installer_id"]
                            )
                            if not installer:
                                installer = EveCharacter.objects.create_character(
                                    job["installer_id"]
                                )
                            IndustryJob.objects.create(
                                id=job["job_id"],
                                activity=job["activity_id"],
                                owner=self,
                                location=self._fetch_location(
                                    job["output_location_id"],
                                    token=token,
                                ),
                                blueprint=Blueprint.objects.get(pk=job["blueprint_id"]),
                                installer=installer,
                                runs=job["runs"],
                                start_date=job["start_date"],
                                end_date=job["end_date"],
                                status=job["status"],
                            )
                else:
                    blueprint_id = job["blueprint_id"]
                    logger.warning(f"Unmatchable blueprint ID: {blueprint_id}")
            IndustryJob.objects.filter(pk__in=job_ids_to_remove).delete()

    @fetch_token_for_owner(["esi-assets.read_corporation_assets.v1"])
    def _fetch_corporate_assets(self, token) -> list:
        return esi.client.Assets.get_corporations_corporation_id_assets(
            corporation_id=self.corporation.corporation_id,
            token=token.valid_access_token(),
        ).results()

    @fetch_token_for_owner(["esi-assets.read_assets.v1"])
    def _fetch_personal_assets(self, token) -> list:
        return esi.client.Assets.get_characters_character_id_assets(
            character_id=self.character.character.character_id,
            token=token.valid_access_token(),
        ).results()

    @fetch_token_for_owner(["esi-corporations.read_blueprints.v1"])
    def _fetch_corporate_blueprints(self, token) -> list:
        corporation_id = self.corporation.corporation_id

        blueprints = esi.client.Corporation.get_corporations_corporation_id_blueprints(
            corporation_id=corporation_id,
            token=token.valid_access_token(),
        ).results()
        return blueprints

    @fetch_token_for_owner(["esi-characters.read_blueprints.v1"])
    def _fetch_personal_blueprints(self, token) -> list:
        character_id = self.character.character.character_id

        blueprints = esi.client.Character.get_characters_character_id_blueprints(
            character_id=character_id,
            token=token.valid_access_token(),
        ).results()
        return blueprints

    @fetch_token_for_owner(["esi-industry.read_corporation_jobs.v1"])
    def _fetch_corporate_industry_jobs(self, token) -> list:
        corporation_id = self.corporation.corporation_id

        jobs = esi.client.Industry.get_corporations_corporation_id_industry_jobs(
            corporation_id=corporation_id,
            token=token.valid_access_token(),
        ).results()
        return jobs

    @fetch_token_for_owner(["esi-industry.read_character_jobs.v1"])
    def _fetch_personal_industry_jobs(self, token) -> list:
        character_id = self.character.character.character_id

        jobs = esi.client.Industry.get_characters_character_id_industry_jobs(
            character_id=character_id,
            token=token.valid_access_token(),
        ).results()
        return jobs

    def token(self, scopes=None) -> Tuple[Token, int]:
        """returns a valid Token for the owner"""
        token = None
        error = None
        add_prefix = self._logger_prefix()

        # abort if character is not configured
        if self.character is None:
            logger.error(add_prefix("No character configured to sync"))
            error = self.ERROR_NO_CHARACTER

        # abort if character does not have sufficient permissions
        elif self.corporation and not self.character.user.has_perm(
            "blueprints.add_corporate_blueprint_owner"
        ):
            logger.error(
                add_prefix(
                    "This character does not have sufficient permission to sync corporations"
                )
            )
            error = self.ERROR_INSUFFICIENT_PERMISSIONS

        # abort if character does not have sufficient permissions
        elif not self.character.user.has_perm(
            "blueprints.add_personal_blueprint_owner"
        ):
            logger.error(
                add_prefix(
                    "This character does not have sufficient permission to sync personal blueprints"
                )
            )
            error = self.ERROR_INSUFFICIENT_PERMISSIONS

        else:
            try:
                # get token
                token = (
                    Token.objects.filter(
                        user=self.character.user,
                        character_id=self.character.character.character_id,
                    )
                    .require_scopes(scopes)
                    .require_valid()
                    .first()
                )
            except TokenInvalidError:
                logger.error(add_prefix("Invalid token for fetching blueprints"))
                error = self.ERROR_TOKEN_INVALID
            except TokenExpiredError:
                logger.error(add_prefix("Token expired for fetching blueprints"))
                error = self.ERROR_TOKEN_EXPIRED
            else:
                if not token:
                    logger.error(add_prefix("No token found with sufficient scopes"))
                    error = self.ERROR_TOKEN_INVALID

        return token, error

    def _fetch_location(self, location_id, token) -> "Location":
        return Location.objects.get_or_create_esi_async(id=location_id, token=token)[0]

    def _logger_prefix(self):
        """returns standard logger prefix function"""
        if self.corporation:
            return make_logger_prefix(self.corporation.corporation_ticker)
        return make_logger_prefix(self.character.character.character_name)


class Blueprint(models.Model):
    item_id = models.PositiveBigIntegerField(
        primary_key=True, help_text="The EVE Item ID of the blueprint"
    )
    owner = models.ForeignKey(
        Owner,
        on_delete=models.CASCADE,
        help_text="Corporation that owns the blueprint",
        # TODO: With next model change: Add related name
    )
    eve_type = models.ForeignKey(
        EveType,
        on_delete=models.CASCADE,
        help_text="Blueprint type"
        # TODO: with next model change: remove related name
    )
    location = models.ForeignKey(
        "Location", on_delete=models.CASCADE, help_text="Blueprint location"
    )
    _location_flag_choices = []
    for choice in EVE_LOCATION_FLAGS:
        _location_flag_choices.append((choice, choice))

    location_flag = models.CharField(
        help_text="Additional location information",
        choices=_location_flag_choices,
        max_length=36,
    )
    quantity = models.PositiveIntegerField(help_text="Number of blueprints", default=1)
    runs = models.PositiveIntegerField(
        blank=True,
        null=True,
        help_text="Runs remaining or null if the blueprint is an original",
    )
    material_efficiency = models.PositiveIntegerField(
        help_text="Material efficiency of the blueprint",
        validators=[validate_material_efficiency],
    )
    time_efficiency = models.PositiveIntegerField(
        help_text="Time efficiency of the blueprint",
        validators=[validate_time_efficiency],
    )

    objects = BlueprintManager()

    @property
    def is_original(self):
        return not self.runs

    class Meta:
        default_permissions = ()

    def __str__(self):
        return (
            self.eve_type.name + f" ({self.material_efficiency}/{self.time_efficiency})"
        )

    def has_industryjob(self):
        return hasattr(self, "industryjob") and self.industryjob is not None


class IndustryJob(models.Model):
    id = models.PositiveBigIntegerField(
        primary_key=True,
        help_text=("Eve Online job ID"),
    )

    class Activity(models.IntegerChoices):
        MANUFACTURING = 1, _("Manufacturing")
        RESEARCHING_TECHNOLOGY = 2, _("Researching Technology")
        RESEARCHING_TIME_EFFICIENCY = 3, _("Researching Time Efficiency")
        RESEARCHING_MATERIAL_EFFICIENCY = 4, _("Researching Material Efficiency")
        COPYING = 5, _("Copying")
        DUPLICATING = 6, _("Duplicating")
        REVERSE_ENGINEERING = 7, _("Reverse Engineering")
        INVENTING = 8, _("Inventing")
        REACTING = 9, _("Reacting")

    activity = models.PositiveIntegerField(choices=Activity.choices)
    location = models.ForeignKey(
        "Location",
        on_delete=models.CASCADE,
        help_text=(
            "Eve Online location ID of the facility in which the job is running"
        ),
        related_name="+",
    )
    blueprint = models.OneToOneField(
        Blueprint,
        on_delete=models.CASCADE,
        help_text=("Blueprint the job is running"),
    )
    installer = models.ForeignKey(
        EveCharacter,
        on_delete=models.CASCADE,
        related_name="jobs",
    )
    owner = models.ForeignKey(
        Owner,
        on_delete=models.CASCADE,
        related_name="jobs",
    )
    start_date = models.DateTimeField()
    end_date = models.DateTimeField()
    runs = models.PositiveIntegerField()
    status = models.CharField(
        choices=(
            ("active", "Active"),
            ("cancelled", "Cancelled"),
            ("delivered", "Delivered"),
            ("paused", "Paused"),
            ("ready", "Ready"),
            ("reverted", "Reverted"),
        ),
        max_length=10,
    )


class Location(models.Model):
    """An Eve Online location: Station or Upwell Structure or Solar System"""

    _SOLAR_SYSTEM_ID_START = 30_000_000
    _SOLAR_SYSTEM_ID_END = 33_000_000
    _STATION_ID_START = 60_000_000
    _STATION_ID_END = 64_000_000
    _STRUCTURE_ID_START = 1_000_000_000_000

    id = models.PositiveBigIntegerField(
        primary_key=True,
        help_text=(
            "Eve Online location ID, "
            "either item ID for stations or structure ID for structures"
        ),
    )
    parent = models.ForeignKey(
        "Location",
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        help_text=("Eve Online location ID of the parent object"),
        related_name="+",
    )

    name = models.CharField(
        max_length=NAMES_MAX_LENGTH,
        help_text="In-game name of this station or structure",
    )
    eve_solar_system = models.ForeignKey(
        EveSolarSystem,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
    )
    eve_type = models.ForeignKey(
        EveType,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
    )
    owner = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        help_text="corporation this station or structure belongs to",
        related_name="+",
    )
    updated_at = models.DateTimeField(auto_now=True)

    objects = LocationManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        if self.name:
            return self.name
        if self.eve_type:
            return str(self.eve_type)
        return f"Location #{self.id}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, name='{self.name}')"

    @property
    def is_empty(self) -> bool:
        return not self.eve_solar_system and not self.eve_type and not self.parent_id

    @property
    def solar_system_url(self) -> str:
        """returns dotlan URL for this solar system"""
        try:
            return dotlan.solar_system_url(self.eve_solar_system.name)
        except AttributeError:
            return ""

    @property
    def is_solar_system(self) -> bool:
        return self.is_solar_system_id(self.id)

    @property
    def is_station(self) -> bool:
        return self.is_station_id(self.id)

    @property
    def is_structure(self) -> bool:
        return self.is_structure_id(self.id)

    @classmethod
    def is_solar_system_id(cls, location_id: int) -> bool:
        return cls._SOLAR_SYSTEM_ID_START <= location_id <= cls._SOLAR_SYSTEM_ID_END

    @classmethod
    def is_station_id(cls, location_id: int) -> bool:
        return cls._STATION_ID_START <= location_id <= cls._STATION_ID_END

    @classmethod
    def is_structure_id(cls, location_id: int) -> bool:
        return location_id >= cls._STRUCTURE_ID_START

    def full_qualified_name(self) -> str:
        """Return the full qualified name of this location."""
        if self.parent:
            return f"{self.parent.full_qualified_name()} - {str(self)}"
        return str(self)


class Request(models.Model):
    blueprint = models.ForeignKey(
        Blueprint,
        on_delete=models.CASCADE,
    )
    requesting_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name="+",
        help_text="The requesting user",
    )
    fulfulling_user = models.ForeignKey(
        User,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="+",
        help_text="The user that fulfilled the request, if it has been fulfilled",
    )
    runs = models.PositiveIntegerField(
        blank=True,
        null=True,
        help_text="Runs requested or blank for maximum allowed",
    )
    STATUS_OPEN = "OP"
    STATUS_IN_PROGRESS = "IP"
    STATUS_FULFILLED = "FL"
    STATUS_CANCELLED = "CL"

    STATUS_CHOICES = [
        (STATUS_OPEN, "Open"),
        (STATUS_IN_PROGRESS, "In Progress"),
        (STATUS_FULFILLED, "Fulfilled"),
        (STATUS_CANCELLED, "Cancelled"),
    ]
    status = models.CharField(
        help_text="Status of the blueprint request",
        choices=STATUS_CHOICES,
        max_length=2,
        db_index=True,
    )
    created_at = models.DateTimeField(auto_now_add=True, db_index=True)
    closed_at = models.DateTimeField(blank=True, null=True, db_index=True)

    objects = RequestManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return f"{self.requesting_user.profile.main_character.character_name}'s request for {self.blueprint.eve_type.name}"

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id={self.pk}, "
            f"requesting_user='{self.requesting_user.profile.main_character.character_name}', "
            f"type_name='{self.eve_type.name}')"
        )

    def notify_new_request(self) -> None:
        for approver in self.approvers():
            notify(
                title=(
                    f"Blueprints: {self.blueprint.eve_type.name} has been requested"
                ),
                message=(
                    f"A copy of {self.blueprint.eve_type.name} has been "
                    f"requested by {self.requesting_user}."
                ),
                user=approver,
                level="info",
            )

    def notify_request_in_progress(self) -> None:
        notify(
            title=(f"Blueprints: {self.blueprint.eve_type.name} request in progress"),
            message=(
                f"{self.fulfulling_user} has started producing copies for "
                f"{self.blueprint.eve_type.name}."
            ),
            user=self.requesting_user,
            level="info",
        )
        other_approvers = set(self.approvers()) - {self.fulfulling_user}
        for approver in other_approvers:
            notify(
                title=(
                    f"Blueprints: {self.blueprint.eve_type.name} request in progress"
                ),
                message=(
                    f"{self.fulfulling_user} has started producing copies for "
                    f"{self.blueprint.eve_type.name} as requested by"
                    f"{self.requesting_user} "
                ),
                user=approver,
                level="info",
            )

    def notify_request_reopened(self, reopened_by: User) -> None:
        for user in set(self.approvers()) - {reopened_by} | {self.requesting_user}:
            notify(
                title=(f"Blueprints: {self.blueprint.eve_type.name} request re-opened"),
                message=(
                    f"{reopened_by} has re-opened the request for "
                    f"{self.blueprint.eve_type.name} by"
                    f"{self.requesting_user}"
                ),
                user=user,
                level="warning",
            )

    def notify_request_fulfilled(self) -> None:
        notify(
            title=(f"Blueprints: {self.blueprint.eve_type.name} request completed"),
            message=(
                f"{self.fulfulling_user} has finished producing copies for "
                f"{self.blueprint.eve_type.name}."
            ),
            user=self.requesting_user,
            level="success",
        )

    def notify_request_canceled_by_requestor(self) -> None:
        for approver in set(self.approvers()):
            notify(
                title=(f"Blueprints: {self.blueprint.eve_type.name} request canceled"),
                message=(
                    f"{self.requesting_user} has canceled his request for "
                    f"{self.blueprint.eve_type.name}."
                ),
                user=approver,
                level="danger",
            )

    def notify_request_canceled_by_approver(self, canceled_by: User) -> None:
        for approver in set(self.approvers()) - {canceled_by} | {self.requesting_user}:
            notify(
                title=(f"Blueprints: {self.blueprint.eve_type.name} request canceled"),
                message=(
                    f"{canceled_by} has canceled the request for "
                    f"{self.blueprint.eve_type.name} "
                    f"by {self.requesting_user}."
                ),
                user=approver,
                level="danger",
            )

    @classmethod
    def approvers(cls) -> List[User]:
        permission = Permission.objects.select_related("content_type").get(
            content_type__app_label=cls._meta.app_label, codename="manage_requests"
        )
        return users_with_permission(permission)



from datetime import datetime

from bravado.exception import HTTPNotFound

from django.test import TestCase

from .main import EsiClientStub, EsiEndpoint

testdata = {
    "Alpha": {
        "get_cake": {"1": "cheesecake", "2": "strawberrycake"},
        "get_details": {"1": {"appointment": "2015-03-24T11:37:00Z"}},
        "get_secret": {"1": "blue secret", "2": "red secret"},
        "get_simple": "steak",
        "get_double_impact": {"1": {"A": "Madrid", "B": "Tokyo"}},
    }
}


class TestEsiClientStub(TestCase):
    def setUp(self) -> None:
        self.stub = EsiClientStub(
            testdata,
            [
                EsiEndpoint("Alpha", "get_cake", "cake_id"),
                EsiEndpoint("Alpha", "get_secret", "secret_id", needs_token=True),
                EsiEndpoint("Alpha", "get_simple"),
                EsiEndpoint("Alpha", "get_double_impact", ("first_id", "second_id")),
            ],
        )

    def test_can_create_endpoint(self):
        self.assertTrue(hasattr(self.stub, "Alpha"))
        self.assertTrue(hasattr(self.stub.Alpha, "get_cake"))
        self.assertEqual(self.stub.Alpha.get_cake(cake_id=1).results(), "cheesecake")
        self.assertEqual(
            self.stub.Alpha.get_cake(cake_id=2).results(), "strawberrycake"
        )
        self.assertEqual(self.stub.Alpha.get_simple().results(), "steak")
        self.assertEqual(
            self.stub.Alpha.get_double_impact(first_id="1", second_id="B").results(),
            "Tokyo",
        )

    def test_raises_exception_on_wrong_pk(self):
        with self.assertRaises(ValueError):
            self.stub.Alpha.get_cake(fruit_id=1).results()

    def test_raises_exception_on_missing_pk(self):
        with self.assertRaises(ValueError):
            self.stub.Alpha.get_cake().results()

    def test_raises_exception_on_missing_data(self):
        with self.assertRaises(HTTPNotFound):
            self.stub.Alpha.get_cake(cake_id=3).results()

    def test_raises_exception_on_missing_token_if_required(self):
        with self.assertRaises(ValueError):
            self.stub.Alpha.get_secret(secret_id=1).results()

    def test_raises_exception_when_trying_to_refine_same_endpoint(self):
        with self.assertRaises(ValueError):
            EsiClientStub(
                testdata,
                [
                    EsiEndpoint("Alpha", "get_cake", "cake_id"),
                    EsiEndpoint("Alpha", "get_cake", "cake_id"),
                ],
            )

    def test_raises_exception_when_trying_to_refine_endpoint_without_data(self):
        with self.assertRaises(ValueError):
            EsiClientStub(
                testdata,
                [
                    EsiEndpoint("Alpha", "get_fruit_id", "fruit_id"),
                ],
            )

    def test_can_convert_datetimes(self):
        stub = EsiClientStub(
            testdata,
            [
                EsiEndpoint("Alpha", "get_details", "id"),
            ],
        )
        results = stub.Alpha.get_details(id=1).results()
        self.assertIsInstance(results["appointment"], datetime)



from typing import Generic, TypeVar

import factory
import factory.fuzzy

from eveuniverse.models import EveSolarSystem, EveType

from blueprints import constants
from blueprints.models import Blueprint, Location, Owner

T = TypeVar("T")

faker = factory.faker.faker.Faker()


class BaseMetaFactory(Generic[T], factory.base.FactoryMetaClass):
    def __call__(cls, *args, **kwargs) -> T:
        return super().__call__(*args, **kwargs)


class LocationStationFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[Location]
):
    class Meta:
        model = Location

    id = factory.Sequence(lambda o: o + 60_000_000)
    name = factory.Faker("city")

    @factory.lazy_attribute
    def eve_solar_system(self):
        return EveSolarSystem.objects.order_by("?").first()

    @factory.lazy_attribute
    def eve_type(self):
        return (
            EveType.objects.filter(
                eve_group__eve_category_id=constants.EVE_CATEGORY_ID_STATION,
                published=True,
            )
            .order_by("?")
            .first()
        )


class OwnerFactory(factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[Owner]):
    class Meta:
        model = Owner


class BlueprintFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[Blueprint]
):
    class Meta:
        model = Blueprint

    item_id = factory.Sequence(lambda o: o + 100_000_000)
    owner = factory.SubFactory(OwnerFactory)
    location = factory.SubFactory(LocationStationFactory)
    location_flag = "Hangar"
    quantity = 1
    runs = None
    material_efficiency = 10
    time_efficiency = 20

    @factory.lazy_attribute
    def eve_type(self):
        return (
            EveType.objects.filter(
                eve_group__eve_category_id=constants.EVE_CATEGORY_ID_BLUEPRINT,
                published=True,
            )
            .order_by("?")
            .first()
        )



"""Settings for Blueprints."""

from app_utils.app_settings import clean_setting

BLUEPRINTS_ESI_TIMEOUT_ENABLED = clean_setting("BLUEPRINTS_ESI_TIMEOUT_ENABLED", True)

BLUEPRINTS_ESI_ERROR_LIMIT_THRESHOLD = clean_setting(
    "BLUEPRINTS_ESI_ERROR_LIMIT_THRESHOLD", 25
)

BLUEPRINTS_ADMIN_NOTIFICATIONS_ENABLED = clean_setting(
    "BLUEPRINTS_ADMIN_NOTIFICATIONS_ENABLED", True
)

# Hard timeout for tasks in seconds to reduce task accumulation during outages
BLUEPRINTS_TASKS_TIME_LIMIT = clean_setting("BLUEPRINTS_TASKS_TIME_LIMIT", 7200)

# Default page size for BLUEPRINT list.
# Must be an integer value from the current options as seen in the app.
BLUEPRINTS_DEFAULT_PAGE_LENGTH = clean_setting("BLUEPRINTS_DEFAULT_PAGE_LENGTH", 10)

# Wether paging is enabled for the BLUEPRINT list
BLUEPRINTS_PAGING_ENABLED = clean_setting("BLUEPRINTS_PAGING_ENABLED", True)
BLUEPRINTS_LIST_ICON_OUTPUT_SIZE = clean_setting("BLUEPRINTS_LIST_ICON_OUTPUT_SIZE", 32)

# Hours after a existing location (e.g. structure) becomes stale and gets updated
# e.g. for name changes of structures
BLUEPRINTS_LOCATION_STALE_HOURS = clean_setting("BLUEPRINTS_LOCATION_STALE_HOURS", 24)



"""Admin site for Blueprints."""

from typing import Any

from django.contrib import admin
from django.db.models.query import QuerySet
from django.http.request import HttpRequest
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from .models import Blueprint, IndustryJob, Location, Owner, Request


class LocationNameListFilter(admin.SimpleListFilter):
    title = _("location name")
    parameter_name = "location_name"

    def lookups(self, request, model_admin):
        qs = model_admin.get_queryset(request)
        return (
            (obj, obj)
            for obj in qs.values_list("location_name", flat=True)
            .distinct()
            .order_by("location_name")
        )

    def queryset(self, request, queryset):
        if value := self.value():
            return queryset.filter(location_name=value)


class BpoListFilter(admin.SimpleListFilter):
    title = _("is BPO")
    parameter_name = "is_bpo"

    def lookups(self, request, model_admin):
        return (
            ("yes", _("yes")),
            ("no", _("no")),
        )

    def queryset(self, request, queryset):
        if value := self.value():
            return queryset.filter(is_bpo=value)


class LocationFlagListFilter(admin.SimpleListFilter):
    title = _("location flag")
    parameter_name = "location_flag"

    def lookups(self, request, model_admin):
        qs = model_admin.get_queryset(request)
        return (
            (obj, obj)
            for obj in qs.values_list("location_flag", flat=True)
            .distinct()
            .order_by("location_flag")
        )

    def queryset(self, request, queryset):
        if value := self.value():
            return queryset.filter(location_flag=value)


@admin.register(Blueprint)
class BlueprintAdmin(admin.ModelAdmin):
    list_display = (
        "_type",
        "_owner",
        "_location_name",
        "location_flag",
        "_material_efficiency",
        "_time_efficiency",
        "_original",
    )
    list_filter = [
        "owner",
        BpoListFilter,
        LocationNameListFilter,
        LocationFlagListFilter,
        ("eve_type__eve_group", admin.RelatedOnlyFieldListFilter),
    ]
    search_fields = ["eve_type__name"]

    def get_queryset(self, request: HttpRequest) -> QuerySet[Any]:
        qs = (
            super()
            .get_queryset(request)
            .annotate_is_bpo()
            .select_related("eve_type", "owner", "owner__corporation")
        )
        return qs.annotate_location_name()

    @admin.display(ordering="eve_type__name")
    def _type(self, obj):
        return obj.eve_type.name if obj.eve_type else None

    @admin.display(ordering="owner__name")
    def _owner(self, obj):
        return obj.owner.name

    @admin.display(ordering="is_bpo", boolean=True, description="BPO")
    def _original(self, obj) -> bool:
        return obj.is_bpo == "yes"

    def _location_name(self, obj):
        return obj.location_name

    @admin.display(description="ME", ordering="material_efficiency")
    def _material_efficiency(self, obj):
        return obj.material_efficiency

    @admin.display(description="TE", ordering="time_efficiency")
    def _time_efficiency(self, obj):
        return obj.time_efficiency

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(Location)
class LocationAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "name",
        "_type",
        "_group",
        "_solar_system",
        "_parent",
        "updated_at",
    )
    list_filter = (
        # ("parent", admin.RelatedOnlyFieldListFilter),
        (
            "eve_solar_system__eve_constellation__eve_region",
            admin.RelatedOnlyFieldListFilter,
        ),
        ("eve_solar_system", admin.RelatedOnlyFieldListFilter),
        ("eve_type__eve_group", admin.RelatedOnlyFieldListFilter),
    )
    search_fields = ["name"]

    def get_queryset(self, request: HttpRequest) -> QuerySet[Any]:
        qs = super().get_queryset(request)
        return qs.select_related(
            "eve_solar_system",
            "eve_solar_system__eve_constellation__eve_region",
            "eve_type",
            "eve_type__eve_group",
            "parent",
            "parent__eve_type",
        )

    @admin.display(ordering="eve_solar_system__name")
    def _solar_system(self, obj: Location):
        return obj.eve_solar_system.name if obj.eve_solar_system else None

    @admin.display(ordering="eve_type__name")
    def _type(self, obj: Location):
        return obj.eve_type.name if obj.eve_type else None

    @admin.display(ordering="eve_type__eve_group__name")
    def _group(self, obj: Location):
        return obj.eve_type.eve_group.name if obj.eve_type else None

    def _parent(self, obj: Location):
        if not obj.parent:
            return None
        url = reverse("admin:blueprints_location_change", args=(obj.parent.id,))
        return format_html('<a href="{}">{}</a>', url, obj.parent)

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(Owner)
class OwnerAdmin(admin.ModelAdmin):
    list_display = ("character", "_type", "corporation", "is_active")

    def _type(self, obj):
        return "Corporate" if obj.corporation else "Personal"

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(Request)
class RequestAdmin(admin.ModelAdmin):
    list_display = ("_type", "_requestor", "_owner", "_fulfilled_by")

    list_select_related = (
        "blueprint__eve_type",
        "requesting_user__profile__main_character",
    )
    search_fields = ["blueprint__eve_type__name"]

    def _type(self, obj):
        return obj.blueprint.eve_type.name if obj.blueprint.eve_type else None

    def _requestor(self, obj):
        return obj.requesting_user.profile.main_character.character_name

    def _owner(self, obj):
        return obj.blueprint.owner.name

    def _fulfilled_by(self, obj):
        return (
            obj.fulfulling_user.profile.main_character.character_name
            if obj.fulfulling_user
            else None
        )

    def has_add_permission(self, request):
        return False


@admin.register(IndustryJob)
class IndustryJobAdmin(admin.ModelAdmin):
    list_display = ("_blueprint", "_installer", "_activity")

    list_select_related = ("blueprint__eve_type",)
    search_fields = ["blueprint__eve_type__name"]

    def _blueprint(self, obj):
        return obj.blueprint.eve_type.name if obj.blueprint.eve_type else None

    def _installer(self, obj):
        return obj.installer.character_name

    def _activity(self, obj):
        return obj.get_activity_display()

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False



# Generated by Django 3.1.10 on 2021-05-10 11:41

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("blueprints", "0002_add_locations_permission"),
    ]

    operations = [
        migrations.AlterField(
            model_name="request",
            name="status",
            field=models.CharField(
                choices=[
                    ("OP", "Open"),
                    ("IP", "In Progress"),
                    ("FL", "Fulfilled"),
                    ("CL", "Cancelled"),
                ],
                db_index=True,
                help_text="Status of the blueprint request",
                max_length=2,
            ),
        ),
    ]



"""Validators for Blueprints."""

from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


def validate_material_efficiency(value):
    if value > 10 or value < 0:
        raise ValidationError(
            _("%(value)s is not a valid material efficiency"),
            params={"value": value},
        )


def validate_time_efficiency(value):
    if value % 2 != 0 or value > 20 or value < 0:
        raise ValidationError(
            _("%(value)s is not a valid time efficiency"),
            params={"value": value},
        )
